@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2019 Alexander Mikhaylenko <exalm7659@gmail.com>
+ * Copyright (C) 2019 Alice Mikhaylenko <alicem@gnome.org>
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  */
@@ -9,6 +9,7 @@
 #include "adw-carousel.h"
 
 #include "adw-animation-util.h"
+#include "adw-marshalers.h"
 #include "adw-navigation-direction.h"
 #include "adw-spring-animation.h"
 #include "adw-swipe-tracker.h"
@@ -77,6 +78,7 @@ struct _AdwCarousel
 
   guint scroll_timeout_id;
   gboolean can_scroll;
+  gboolean is_being_allocated;
 };
 
 static void adw_carousel_buildable_init (GtkBuildableIface *iface);
@@ -594,7 +596,7 @@ adw_carousel_size_allocate (GtkWidget *widget,
   gboolean is_rtl;
   double snap_point;
 
-  if (self->position_shift != 0) {
+  if (!G_APPROX_VALUE (self->position_shift, 0, DBL_EPSILON)) {
     set_position (self, self->position + self->position_shift);
     adw_swipe_tracker_shift_position (self->tracker, self->position_shift);
     self->position_shift = 0;
@@ -653,9 +655,6 @@ adw_carousel_size_allocate (GtkWidget *widget,
                                          child_info->snap_point);
   }
 
-  if (!gtk_widget_get_realized (GTK_WIDGET (self)))
-    return;
-
   x = 0;
   y = 0;
 
@@ -705,6 +704,8 @@ adw_carousel_size_allocate (GtkWidget *widget,
     else
       x += self->distance * child_info->size;
   }
+
+  self->is_being_allocated = FALSE;
 }
 
 static void
@@ -1011,17 +1012,22 @@ adw_carousel_class_init (AdwCarouselClass *klass)
    * This signal is emitted after a page has been changed.
    *
    * It can be used to implement "infinite scrolling" by amending the pages
-   * after every scroll.
+   * after every scroll. Note that an empty carousel is indicated by
+   * `(int)index == -1`.
    */
   signals[SIGNAL_PAGE_CHANGED] =
     g_signal_new ("page-changed",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  NULL, NULL, NULL,
+                  NULL, NULL,
+                  adw_marshal_VOID__UINT,
                   G_TYPE_NONE,
                   1,
                   G_TYPE_UINT);
+  g_signal_set_va_marshaller (signals[SIGNAL_PAGE_CHANGED],
+                              G_TYPE_FROM_CLASS (klass),
+                              adw_marshal_VOID__UINTv);
 
   gtk_widget_class_set_css_name (widget_class, "carousel");
 }
@@ -1168,6 +1174,7 @@ adw_carousel_prepend (AdwCarousel *self,
 {
   g_return_if_fail (ADW_IS_CAROUSEL (self));
   g_return_if_fail (GTK_IS_WIDGET (widget));
+  g_return_if_fail (gtk_widget_get_parent (widget) == NULL);
 
   adw_carousel_insert (self, widget, 0);
 }
@@ -1185,6 +1192,7 @@ adw_carousel_append (AdwCarousel *self,
 {
   g_return_if_fail (ADW_IS_CAROUSEL (self));
   g_return_if_fail (GTK_IS_WIDGET (widget));
+  g_return_if_fail (gtk_widget_get_parent (widget) == NULL);
 
   adw_carousel_insert (self, widget, -1);
 }
@@ -1210,6 +1218,7 @@ adw_carousel_insert (AdwCarousel *self,
 
   g_return_if_fail (ADW_IS_CAROUSEL (self));
   g_return_if_fail (GTK_IS_WIDGET (widget));
+  g_return_if_fail (gtk_widget_get_parent (widget) == NULL);
   g_return_if_fail (position >= -1);
 
   info = g_new0 (ChildInfo, 1);
@@ -1230,6 +1239,7 @@ adw_carousel_insert (AdwCarousel *self,
     gtk_widget_set_parent (widget, GTK_WIDGET (self));
   }
 
+  self->is_being_allocated = TRUE;
   gtk_widget_queue_allocate (GTK_WIDGET (self));
 
   animate_child_resize (self, info, 1, self->reveal_duration);
@@ -1316,11 +1326,13 @@ adw_carousel_reorder (AdwCarousel *self,
     gtk_widget_insert_before (child, GTK_WIDGET (self), NULL);
   }
 
-  if (closest_point == old_point)
+  if (G_APPROX_VALUE (closest_point, old_point, DBL_EPSILON))
     self->position_shift += new_point - old_point;
-  else if (old_point >= closest_point && closest_point >= new_point)
+  else if ((G_APPROX_VALUE (old_point, closest_point, DBL_EPSILON) || old_point > closest_point) &&
+           (G_APPROX_VALUE (closest_point, new_point, DBL_EPSILON) || closest_point > new_point))
     self->position_shift += info->size;
-  else if (new_point >= closest_point && closest_point >= old_point)
+  else if ((G_APPROX_VALUE (new_point, closest_point, DBL_EPSILON) || new_point > closest_point) &&
+           (G_APPROX_VALUE (closest_point, old_point, DBL_EPSILON) || closest_point > old_point))
     self->position_shift -= info->size;
 
   gtk_widget_queue_allocate (GTK_WIDGET (self));
@@ -1359,6 +1371,35 @@ adw_carousel_remove (AdwCarousel *self,
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_N_PAGES]);
 }
 
+static void
+do_scroll_to (AdwCarousel *self,
+              GtkWidget   *widget,
+              gboolean     animate)
+{
+  scroll_to (self, widget, 0);
+
+  if (!animate)
+    adw_animation_skip (self->animation);
+}
+
+typedef struct {
+  AdwCarousel *carousel;
+  GtkWidget *widget;
+  gboolean animate;
+} ScrollData;
+
+static gboolean
+scroll_to_idle_cb (ScrollData *data)
+{
+  do_scroll_to (data->carousel, data->widget, data->animate);
+
+  g_object_unref (data->carousel);
+  g_object_unref (data->widget);
+  g_free (data);
+
+  return FALSE;
+}
+
 /**
  * adw_carousel_scroll_to:
  * @self: a carousel
@@ -1378,10 +1419,24 @@ adw_carousel_scroll_to (AdwCarousel *self,
   g_return_if_fail (GTK_IS_WIDGET (widget));
   g_return_if_fail (gtk_widget_get_parent (widget) == GTK_WIDGET (self));
 
-  scroll_to (self, widget, 0);
+  if (self->is_being_allocated) {
+    /* 'self' is still being allocated/resized by GTK
+     * async machinery (initiated from a previous adw_carousel_insert() call)
+     * So in this case let's do the scroll in an idle handler so
+     * it gets executed when everything is in place i.e. after GTK
+     * calls our adw_carousel_size_allocate() function - issue #597 */
+    ScrollData *data;
 
-  if (!animate)
-    adw_animation_skip (self->animation);
+    data = g_new (ScrollData, 1);
+    data->carousel = g_object_ref (self);
+    data->widget = g_object_ref (widget);
+    data->animate = animate;
+
+    g_idle_add (G_SOURCE_FUNC (scroll_to_idle_cb), data);
+    return;
+  }
+
+  do_scroll_to (self, widget, animate);
 }
 
 /**
