@@ -6,7 +6,7 @@
 
 /*
  * Heavily based on GTK 3.99.3 GtkDropDown widget initially written by Matthias
- * Clasen, and heavily modified for libadwaita by Alexander Mikhaylenko on
+ * Clasen, and heavily modified for libadwaita by Alice Mikhaylenko on
  * behalf of Purism SPC 2020.
  */
 
@@ -63,8 +63,6 @@
  * ## Accessibility
  *
  * `AdwComboRow` uses the `GTK_ACCESSIBLE_ROLE_COMBO_BOX` role.
- *
- * Since: 1.0
  */
 
 /*
@@ -77,11 +75,14 @@ typedef struct
   GtkListView *current;
   GtkListView *list;
   GtkPopover *popover;
+  GtkSearchEntry *search_entry;
   gboolean use_subtitle;
+  gboolean enable_search;
 
   GtkListItemFactory *factory;
   GtkListItemFactory *list_factory;
   GListModel *model;
+  GListModel *filter_model;
   GtkSelectionModel *selection;
   GtkSelectionModel *popup_selection;
   GtkSelectionModel *current_selection;
@@ -93,13 +94,14 @@ G_DEFINE_TYPE_WITH_PRIVATE (AdwComboRow, adw_combo_row, ADW_TYPE_ACTION_ROW)
 
 enum {
   PROP_0,
-  PROP_MODEL,
   PROP_SELECTED,
   PROP_SELECTED_ITEM,
+  PROP_MODEL,
   PROP_FACTORY,
   PROP_LIST_FACTORY,
   PROP_EXPRESSION,
   PROP_USE_SUBTITLE,
+  PROP_ENABLE_SEARCH,
   LAST_PROP,
 };
 
@@ -183,10 +185,61 @@ notify_popover_visible_cb (AdwComboRow *self)
 {
   AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
 
-  if (gtk_widget_get_visible (GTK_WIDGET (priv->popover)))
+  if (gtk_widget_get_visible (GTK_WIDGET (priv->popover))) {
     gtk_widget_add_css_class (GTK_WIDGET (self), "has-open-popup");
-  else
+  } else {
     gtk_widget_remove_css_class (GTK_WIDGET (self), "has-open-popup");
+    gtk_editable_set_text (GTK_EDITABLE (priv->search_entry), "");
+  }
+}
+
+static void
+update_filter (AdwComboRow *self)
+{
+  AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
+  
+  if (priv->filter_model) {
+      GtkFilter *filter;
+
+      if (priv->expression) 
+        filter = GTK_FILTER (gtk_string_filter_new (gtk_expression_ref (priv->expression)));
+      else 
+        filter = GTK_FILTER (gtk_every_filter_new ());
+      gtk_filter_list_model_set_filter (GTK_FILTER_LIST_MODEL (priv->filter_model), filter);
+      g_object_unref (filter);
+  }
+}
+
+static void
+search_changed_cb (GtkSearchEntry *entry,
+                   AdwComboRow    *self)
+{
+  AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
+  const char *text;
+  GtkFilter *filter;
+
+  text = gtk_editable_get_text (GTK_EDITABLE (entry));
+
+  filter = gtk_filter_list_model_get_filter (GTK_FILTER_LIST_MODEL (priv->filter_model));
+  if (GTK_IS_STRING_FILTER (filter))
+    gtk_string_filter_set_search (GTK_STRING_FILTER (filter), text);
+}
+
+static void
+search_stop_cb (GtkSearchEntry *entry,
+                AdwComboRow    *self)
+{
+  AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
+
+  GtkFilter *filter;
+
+  filter = gtk_filter_list_model_get_filter (GTK_FILTER_LIST_MODEL (priv->filter_model));
+  if (GTK_IS_STRING_FILTER (filter)) {
+    if (gtk_string_filter_get_search (GTK_STRING_FILTER (filter)))
+      gtk_string_filter_set_search (GTK_STRING_FILTER (filter), NULL);
+    else
+      gtk_popover_popdown (GTK_POPOVER (priv->popover));
+  }
 }
 
 static void
@@ -266,12 +319,12 @@ bind_item (GtkSignalListItemFactory *factory,
   }
 
   if (gtk_widget_get_ancestor (box, GTK_TYPE_POPOVER) == GTK_WIDGET (priv->popover)) {
-    gtk_widget_show (icon);
+    gtk_widget_set_visible (icon, TRUE);
     g_signal_connect (self, "notify::selected-item",
                       G_CALLBACK (selected_item_changed), list_item);
     selected_item_changed (self, NULL, list_item);
   } else {
-    gtk_widget_hide (icon);
+    gtk_widget_set_visible (icon, FALSE);
   }
 
   g_clear_pointer (&repr, g_free);
@@ -308,15 +361,15 @@ adw_combo_row_get_property (GObject    *object,
   AdwComboRow *self = ADW_COMBO_ROW (object);
 
   switch (prop_id) {
-  case PROP_MODEL:
-    g_value_set_object (value, adw_combo_row_get_model (self));
-    break;
   case PROP_SELECTED:
     g_value_set_uint (value, adw_combo_row_get_selected (self));
     break;
   case PROP_SELECTED_ITEM:
     g_value_set_object (value, adw_combo_row_get_selected_item (self));
     break;
+  case PROP_MODEL:
+    g_value_set_object (value, adw_combo_row_get_model (self));
+    break;
   case PROP_FACTORY:
     g_value_set_object (value, adw_combo_row_get_factory (self));
     break;
@@ -329,6 +382,9 @@ adw_combo_row_get_property (GObject    *object,
   case PROP_USE_SUBTITLE:
     g_value_set_boolean (value, adw_combo_row_get_use_subtitle (self));
     break;
+  case PROP_ENABLE_SEARCH:
+    g_value_set_boolean (value, adw_combo_row_get_enable_search (self));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -343,12 +399,12 @@ adw_combo_row_set_property (GObject      *object,
   AdwComboRow *self = ADW_COMBO_ROW (object);
 
   switch (prop_id) {
-  case PROP_MODEL:
-    adw_combo_row_set_model (self, g_value_get_object (value));
-    break;
   case PROP_SELECTED:
     adw_combo_row_set_selected (self, g_value_get_uint (value));
     break;
+  case PROP_MODEL:
+    adw_combo_row_set_model (self, g_value_get_object (value));
+    break;
   case PROP_FACTORY:
     adw_combo_row_set_factory (self, g_value_get_object (value));
     break;
@@ -361,6 +417,9 @@ adw_combo_row_set_property (GObject      *object,
   case PROP_USE_SUBTITLE:
     adw_combo_row_set_use_subtitle (self, g_value_get_boolean (value));
     break;
+  case PROP_ENABLE_SEARCH:
+    adw_combo_row_set_enable_search (self, g_value_get_boolean (value));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -386,6 +445,7 @@ adw_combo_row_dispose (GObject *object)
   g_clear_object (&priv->current_selection);
   g_clear_object (&priv->factory);
   g_clear_object (&priv->list_factory);
+  g_clear_object (&priv->filter_model);
 
   g_clear_object (&priv->model);
 
@@ -435,18 +495,6 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
 
   row_class->activate = adw_combo_row_activate;
 
-  /**
-   * AdwComboRow:model: (attributes org.gtk.Property.get=adw_combo_row_get_model org.gtk.Property.set=adw_combo_row_set_model)
-   *
-   * Model for the displayed items.
-   *
-   * Since: 1.0
-   */
-  props[PROP_MODEL] =
-    g_param_spec_object ("model", NULL, NULL,
-                         G_TYPE_LIST_MODEL,
-                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
-
   /**
    * AdwComboRow:selected: (attributes org.gtk.Property.get=adw_combo_row_get_selected org.gtk.Property.set=adw_combo_row_set_selected)
    *
@@ -454,8 +502,6 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
    *
    * If no item is selected, the property has the value
    * [const@Gtk.INVALID_LIST_POSITION]
-   *
-   * Since: 1.0
    */
   props[PROP_SELECTED] =
     g_param_spec_uint ("selected", NULL, NULL,
@@ -466,14 +512,22 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
    * AdwComboRow:selected-item: (attributes org.gtk.Property.get=adw_combo_row_get_selected_item)
    *
    * The selected item.
-   *
-   * Since: 1.0
    */
   props[PROP_SELECTED_ITEM] =
     g_param_spec_object ("selected-item", NULL, NULL,
                          G_TYPE_OBJECT,
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * AdwComboRow:model: (attributes org.gtk.Property.get=adw_combo_row_get_model org.gtk.Property.set=adw_combo_row_set_model)
+   *
+   * The model that provides the displayed items.
+   */
+  props[PROP_MODEL] =
+    g_param_spec_object ("model", NULL, NULL,
+                         G_TYPE_LIST_MODEL,
+                         G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
   /**
    * AdwComboRow:factory: (attributes org.gtk.Property.get=adw_combo_row_get_factory org.gtk.Property.set=adw_combo_row_set_factory)
    *
@@ -481,8 +535,6 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
    *
    * This factory is always used for the item in the row. It is also used for
    * items in the popup unless [property@ComboRow:list-factory] is set.
-   *
-   * Since: 1.0
    */
   props[PROP_FACTORY] =
     g_param_spec_object ("factory", NULL, NULL,
@@ -495,8 +547,6 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
    * The factory for populating list items in the popup.
    *
    * If this is not set, [property@ComboRow:factory] is used.
-   *
-   * Since: 1.0
    */
   props[PROP_LIST_FACTORY] =
     g_param_spec_object ("list-factory", NULL, NULL,
@@ -508,12 +558,11 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
    *
    * An expression used to obtain strings from items.
    *
-   * It's used to bind strings to labels produced by the default factory.
+   * The expression must have a value type of `G_TYPE_STRING`.
    *
-   * If [property@ComboRow:factory] is not set, the expression is also used to
-   * bind strings to labels produced by a default factory.
-   *
-   * Since: 1.0
+   * It's used to bind strings to labels produced by the default factory if
+   * [property@ComboRow:factory] is not set, or when
+   * [property@ComboRow:use-subtitle] is set to `TRUE`.
    */
   props[PROP_EXPRESSION] =
     gtk_param_spec_expression ("expression",
@@ -529,18 +578,33 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
    * If you use a custom list item factory, you will need to give the row a
    * name conversion expression with [property@ComboRow:expression].
    *
-   * If `TRUE`, you should not access [property@ActionRow:subtitle].
+   * If set to `TRUE`, you should not access [property@ActionRow:subtitle].
    *
    * The subtitle is interpreted as Pango markup if
    * [property@PreferencesRow:use-markup] is set to `TRUE`.
-   *
-   * Since: 1.0
    */
   props[PROP_USE_SUBTITLE] =
     g_param_spec_boolean ("use-subtitle", NULL, NULL,
                           FALSE,
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
+  /**
+   * AdwComboRow:enable-search: (attributes org.gtk.Property.get=adw_combo_row_get_enable_search org.gtk.Property.set=adw_combo_row_set_enable_search)
+   *
+   * Whether to show a search entry in the popup.
+   *
+   * If set to `TRUE`, a search entry will be shown in the popup that
+   * allows to search for items in the list.
+   * 
+   * Search requires [property@ComboRow:expression] to be set.
+   * 
+   * Since: 1.4
+   */
+  props[PROP_ENABLE_SEARCH] =
+    g_param_spec_boolean ("enable-search", NULL, NULL,
+                          FALSE,
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
   gtk_widget_class_set_template_from_resource (widget_class,
@@ -549,8 +613,11 @@ adw_combo_row_class_init (AdwComboRowClass *klass)
   gtk_widget_class_bind_template_child_private (widget_class, AdwComboRow, arrow_box);
   gtk_widget_class_bind_template_child_private (widget_class, AdwComboRow, list);
   gtk_widget_class_bind_template_child_private (widget_class, AdwComboRow, popover);
+  gtk_widget_class_bind_template_child_private (widget_class, AdwComboRow, search_entry);
   gtk_widget_class_bind_template_callback (widget_class, row_activated_cb);
   gtk_widget_class_bind_template_callback (widget_class, notify_popover_visible_cb);
+  gtk_widget_class_bind_template_callback (widget_class, search_changed_cb);
+  gtk_widget_class_bind_template_callback (widget_class, search_stop_cb);
 
   gtk_widget_class_set_accessible_role (widget_class, GTK_ACCESSIBLE_ROLE_COMBO_BOX);
 }
@@ -571,8 +638,6 @@ adw_combo_row_init (AdwComboRow *self)
  * Creates a new `AdwComboRow`.
  *
  * Returns: the newly created `AdwComboRow`
- *
- * Since: 1.0
  */
 GtkWidget *
 adw_combo_row_new (void)
@@ -587,8 +652,6 @@ adw_combo_row_new (void)
  *   [const@Gtk.INVALID_LIST_POSITION]
  *
  * Selects the item at the given position.
- *
- * Since: 1.0
  */
 void
 adw_combo_row_set_selected (AdwComboRow *self,
@@ -617,8 +680,6 @@ adw_combo_row_set_selected (AdwComboRow *self,
  *
  * Returns: the position of the selected item, or
  *   [const@Gtk.INVALID_LIST_POSITION] if no item is selected
- *
- * Since: 1.0
  */
 guint
 adw_combo_row_get_selected (AdwComboRow *self)
@@ -642,8 +703,6 @@ adw_combo_row_get_selected (AdwComboRow *self)
  * Gets the selected item.
  *
  * Returns: (transfer none) (type GObject) (nullable): the selected item
- *
- * Since: 1.0
  */
 gpointer
 adw_combo_row_get_selected_item (AdwComboRow *self)
@@ -667,8 +726,6 @@ adw_combo_row_get_selected_item (AdwComboRow *self)
  * Gets the model that provides the displayed items.
  *
  * Returns: (nullable) (transfer none): The model in use
- *
- * Since: 1.0
  */
 GListModel *
 adw_combo_row_get_model (AdwComboRow *self)
@@ -687,9 +744,7 @@ adw_combo_row_get_model (AdwComboRow *self)
  * @self: a combo row
  * @model: (nullable) (transfer none): the model to use
  *
- * Sets the [iface@Gio.ListModel] to use.
- *
- * Since: 1.0
+ * Sets the model that provides the displayed items.
  */
 void
 adw_combo_row_set_model (AdwComboRow *self,
@@ -719,17 +774,25 @@ adw_combo_row_set_model (AdwComboRow *self,
     g_clear_object (&priv->current_selection);
   } else {
     GtkSelectionModel *selection;
+    GListModel *filter_model;
     GListModel *current_model;
 
-    selection = GTK_SELECTION_MODEL (gtk_single_selection_new (g_object_ref (model)));
+    filter_model = G_LIST_MODEL (gtk_filter_list_model_new (g_object_ref (model), NULL));
+    g_set_object (&priv->filter_model, filter_model);
+
+    update_filter (self);
+
+    selection = GTK_SELECTION_MODEL (gtk_single_selection_new (g_object_ref (filter_model)));
     g_set_object (&priv->popup_selection, selection);
     gtk_list_view_set_model (priv->list, selection);
     g_object_unref (selection);
 
-    selection = GTK_SELECTION_MODEL (gtk_single_selection_new (g_object_ref (model)));
+    selection = GTK_SELECTION_MODEL (gtk_single_selection_new (g_object_ref (filter_model)));
     g_set_object (&priv->selection, selection);
     g_object_unref (selection);
 
+    g_object_unref (filter_model);
+
     current_model = G_LIST_MODEL (gtk_selection_filter_model_new (priv->selection));
     selection = GTK_SELECTION_MODEL (gtk_no_selection_new (current_model));
     g_set_object (&priv->current_selection, selection);
@@ -750,11 +813,9 @@ adw_combo_row_set_model (AdwComboRow *self,
  * adw_combo_row_get_factory: (attributes org.gtk.Method.get_property=factory)
  * @self: a combo row
  *
- * Gets the factory that's currently used to populate list items.
+ * Gets the factory for populating list items.
  *
  * Returns: (nullable) (transfer none): the factory in use
- *
- * Since: 1.0
  */
 GtkListItemFactory *
 adw_combo_row_get_factory (AdwComboRow *self)
@@ -773,9 +834,10 @@ adw_combo_row_get_factory (AdwComboRow *self)
  * @self: a combo row
  * @factory: (nullable) (transfer none): the factory to use
  *
- * Sets the `GtkListItemFactory` to use for populating list items.
+ * Sets the factory for populating list items.
  *
- * Since: 1.0
+ * This factory is always used for the item in the row. It is also used for
+ * items in the popup unless [property@ComboRow:list-factory] is set.
  */
 void
 adw_combo_row_set_factory (AdwComboRow        *self,
@@ -803,11 +865,9 @@ adw_combo_row_set_factory (AdwComboRow        *self,
  * adw_combo_row_get_list_factory: (attributes org.gtk.Method.get_property=list-factory)
  * @self: a combo row
  *
- * Gets the factory that's currently used to populate list items in the popup.
+ * Gets the factory for populating list items in the popup.
  *
  * Returns: (nullable) (transfer none): the factory in use
- *
- * Since: 1.0
  */
 GtkListItemFactory *
 adw_combo_row_get_list_factory (AdwComboRow *self)
@@ -826,9 +886,9 @@ adw_combo_row_get_list_factory (AdwComboRow *self)
  * @self: a combo row
  * @factory: (nullable) (transfer none): the factory to use
  *
- * Sets the `GtkListItemFactory` to use for populating list items in the popup.
+ * Sets the factory for populating list items in the popup.
  *
- * Since: 1.0
+ * If this is not set, [property@ComboRow:factory] is used.
  */
 void
 adw_combo_row_set_list_factory (AdwComboRow        *self,
@@ -859,8 +919,6 @@ adw_combo_row_set_list_factory (AdwComboRow        *self,
  * Gets the expression used to obtain strings from items.
  *
  * Returns: (nullable) (transfer none): the expression used to obtain strings from items
- *
- * Since: 1.0
  */
 GtkExpression *
 adw_combo_row_get_expression (AdwComboRow *self)
@@ -883,7 +941,9 @@ adw_combo_row_get_expression (AdwComboRow *self)
  *
  * The expression must have a value type of `G_TYPE_STRING`.
  *
- * Since: 1.0
+ * It's used to bind strings to labels produced by the default factory if
+ * [property@ComboRow:factory] is not set, or when
+ * [property@ComboRow:use-subtitle] is set to `TRUE`.
  */
 void
 adw_combo_row_set_expression (AdwComboRow   *self,
@@ -908,6 +968,8 @@ adw_combo_row_set_expression (AdwComboRow   *self,
 
   selection_changed (self);
 
+  update_filter (self);
+
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_EXPRESSION]);
 }
 
@@ -918,8 +980,6 @@ adw_combo_row_set_expression (AdwComboRow   *self,
  * Gets whether to use the current value as the subtitle.
  *
  * Returns: whether to use the current value as the subtitle
- *
- * Since: 1.0
  */
 gboolean
 adw_combo_row_get_use_subtitle (AdwComboRow *self)
@@ -940,7 +1000,13 @@ adw_combo_row_get_use_subtitle (AdwComboRow *self)
  *
  * Sets whether to use the current value as the subtitle.
  *
- * Since: 1.0
+ * If you use a custom list item factory, you will need to give the row a
+ * name conversion expression with [property@ComboRow:expression].
+ *
+ * If set to `TRUE`, you should not access [property@ActionRow:subtitle].
+ *
+ * The subtitle is interpreted as Pango markup if
+ * [property@PreferencesRow:use-markup] is set to `TRUE`.
  */
 void
 adw_combo_row_set_use_subtitle (AdwComboRow *self,
@@ -964,3 +1030,67 @@ adw_combo_row_set_use_subtitle (AdwComboRow *self,
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_USE_SUBTITLE]);
 }
+
+/**
+ * adw_combo_row_get_enable_search: (attributes org.gtk.Method.set_property=enable-search)
+ * @self: a combo row
+ *
+ * Gets whether search is enabled.
+ * 
+ * If set to `TRUE`, a search entry will be shown in the popup that
+ * allows to search for items in the list.
+ * 
+ * Search requires [property@ComboRow:expression] to be set.
+ *
+ * Returns: whether the popup includes a search entry
+ * 
+ * Since: 1.4
+ */
+gboolean
+adw_combo_row_get_enable_search (AdwComboRow *self)
+{
+  AdwComboRowPrivate *priv;
+
+  g_return_val_if_fail (ADW_IS_COMBO_ROW (self), FALSE);
+
+  priv = adw_combo_row_get_instance_private (self);
+
+  return priv->enable_search;
+}
+
+/**
+ * adw_combo_row_set_enable_search: (attributes org.gtk.Method.set_property=enable-search)
+ * @self: a combo row
+ * @enable_search: whether to enable search
+ *
+ * Sets whether to enable search.
+ * 
+ * If set to `TRUE`, a search entry will be shown in the popup that
+ * allows to search for items in the list.
+ * 
+ * Search requires [property@ComboRow:expression] to be set.
+ * 
+ * Since: 1.4
+ */
+void
+adw_combo_row_set_enable_search (AdwComboRow *self,
+                                 gboolean     enable_search)
+{
+  AdwComboRowPrivate *priv;
+
+  g_return_if_fail (ADW_IS_COMBO_ROW (self));
+
+  priv = adw_combo_row_get_instance_private (self);
+
+  enable_search = !!enable_search;
+
+  if (priv->enable_search == enable_search)
+    return;
+
+  priv->enable_search = enable_search;
+  
+  gtk_editable_set_text (GTK_EDITABLE (priv->search_entry), "");
+  gtk_widget_set_visible (GTK_WIDGET (priv->search_entry), enable_search);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_USE_SUBTITLE]);
+}
