@@ -3,7 +3,7 @@
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  *
- * Author: Alexander Mikhaylenko <alexander.mikhaylenko@puri.sm>
+ * Author: Alice Mikhaylenko <alice.mikhaylenko@puri.sm>
  */
 
 #include "config.h"
@@ -11,6 +11,7 @@
 
 #include "adw-gizmo-private.h"
 #include "adw-gtkbuilder-utils-private.h"
+#include "adw-marshalers.h"
 #include "adw-widget-utils-private.h"
 
 /**
@@ -82,7 +83,7 @@
  * {
  *   const char *response = adw_message_dialog_choose_finish (dialog, result);
  *
- *   ...
+ *   // ...
  * }
  *
  * static void
@@ -204,6 +205,8 @@ typedef struct
   GtkWindow *parent_window;
   int parent_width;
   int parent_height;
+
+  guint parent_state_idle_id;
 } AdwMessageDialogPrivate;
 
 static void adw_message_dialog_buildable_init (GtkBuildableIface *iface);
@@ -254,8 +257,8 @@ static void
 parent_size_cb (AdwMessageDialog *self)
 {
   AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
-  int w = gtk_widget_get_allocated_width (GTK_WIDGET (priv->parent_window));
-  int h = gtk_widget_get_allocated_height (GTK_WIDGET (priv->parent_window));
+  int w = gtk_widget_get_width (GTK_WIDGET (priv->parent_window));
+  int h = gtk_widget_get_height (GTK_WIDGET (priv->parent_window));
 
   if (w == priv->parent_width && h == priv->parent_height)
       return;
@@ -265,6 +268,28 @@ parent_size_cb (AdwMessageDialog *self)
   gtk_widget_queue_resize (GTK_WIDGET (self));
 }
 
+static gboolean
+parent_state_idle_cb (AdwMessageDialog *self)
+{
+  AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
+
+  parent_size_cb (self);
+
+  priv->parent_state_idle_id = 0;
+
+  return G_SOURCE_REMOVE;
+}
+
+static void
+parent_state_cb (AdwMessageDialog *self)
+{
+  AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
+
+  g_clear_handle_id (&priv->parent_state_idle_id, g_source_remove);
+
+  priv->parent_state_idle_id = g_idle_add (G_SOURCE_FUNC (parent_state_idle_cb), self);
+}
+
 static void
 parent_realize_cb (AdwMessageDialog *self)
 {
@@ -281,6 +306,8 @@ parent_realize_cb (AdwMessageDialog *self)
                             G_CALLBACK (parent_size_cb), self);
   g_signal_connect_swapped (surface, "notify::height",
                             G_CALLBACK (parent_size_cb), self);
+  g_signal_connect_swapped (surface, "notify::state",
+                            G_CALLBACK (parent_state_cb), self);
 
   parent_size_cb (self);
 }
@@ -298,6 +325,9 @@ parent_unrealize_cb (AdwMessageDialog *self)
   g_signal_handlers_disconnect_by_func (surface,
                                         G_CALLBACK (parent_size_cb),
                                         self);
+  g_signal_handlers_disconnect_by_func (surface,
+                                        G_CALLBACK (parent_state_cb),
+                                        self);
 
   priv->parent_width = -1;
   priv->parent_height = -1;
@@ -308,6 +338,8 @@ parent_window_notify_cb (AdwMessageDialog *self)
 {
   AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
 
+  g_clear_handle_id (&priv->parent_state_idle_id, g_source_remove);
+
   priv->parent_window = NULL;
   priv->parent_width = -1;
   priv->parent_height = -1;
@@ -323,6 +355,8 @@ set_parent (AdwMessageDialog *self,
     return;
 
   if (priv->parent_window) {
+    g_clear_handle_id (&priv->parent_state_idle_id, g_source_remove);
+
     g_signal_handlers_disconnect_by_func (priv->parent_window,
                                           G_CALLBACK (parent_realize_cb),
                                           self);
@@ -386,6 +420,7 @@ create_response_button (AdwMessageDialog *self,
   GtkWidget *button = gtk_button_new_with_mnemonic (info->label);
 
   gtk_widget_add_css_class (button, "flat");
+  gtk_button_set_can_shrink (GTK_BUTTON (button), TRUE);
 
   switch (info->appearance) {
   case ADW_RESPONSE_SUGGESTED:
@@ -518,10 +553,10 @@ measure_responses_do (AdwMessageDialog *self,
 
     if (horiz == compact) {
       min = MAX (min, child_min);
-      nat = MAX (nat, child_min);
+      nat = MAX (nat, child_nat);
     } else if (horiz) {
       button_min = MAX (button_min, child_min);
-      button_nat = MAX (button_nat, child_min);
+      button_nat = MAX (button_nat, child_nat);
       n_buttons++;
     } else {
       min += child_min;
@@ -534,7 +569,7 @@ measure_responses_do (AdwMessageDialog *self,
 
     if (horiz == compact) {
         min = MAX (min, child_min);
-        nat = MAX (nat, child_min);
+        nat = MAX (nat, child_nat);
       } else {
         min += child_min;
         nat += child_nat;
@@ -574,12 +609,12 @@ measure_responses (GtkWidget      *widget,
     measure_responses_do (self, TRUE, orientation, minimum, NULL);
     measure_responses_do (self, FALSE, orientation, NULL, natural);
   } else {
-    int wide_min = 0;
+    int wide_nat = 0;
 
     if (for_size >= 0)
-      measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, &wide_min, NULL);
+      measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, NULL, &wide_nat);
 
-    measure_responses_do (self, for_size >= 0 && for_size < wide_min,
+    measure_responses_do (self, for_size >= 0 && for_size < wide_nat,
                           orientation, minimum, natural);
   }
 
@@ -598,11 +633,11 @@ allocate_responses (GtkWidget *widget,
   AdwMessageDialog *self = ADW_MESSAGE_DIALOG (gtk_widget_get_root (widget));
   AdwMessageDialogPrivate *priv = adw_message_dialog_get_instance_private (self);
   gboolean compact;
-  int wide_min;
+  int wide_nat;
 
-  measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, &wide_min, NULL);
+  measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, NULL, &wide_nat);
 
-  compact = wide_min > width;
+  compact = wide_nat > width;
 
   if (compact)
     gtk_widget_add_css_class (widget, "compact");
@@ -714,7 +749,7 @@ adw_message_dialog_measure (GtkWidget      *widget,
                                                                NULL, NULL);
 
   if (orientation == GTK_ORIENTATION_HORIZONTAL) {
-    int wide_min, narrow_nat;
+    int wide_nat, narrow_nat;
 
     min_size = MAX (min_size, DIALOG_MIN_WIDTH);
 
@@ -725,12 +760,12 @@ adw_message_dialog_measure (GtkWidget      *widget,
       max_size = DIALOG_MAX_WIDTH;
     }
 
-    measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, &wide_min, NULL);
+    measure_responses_do (self, FALSE, GTK_ORIENTATION_HORIZONTAL, NULL, &wide_nat);
     measure_responses_do (self, TRUE, GTK_ORIENTATION_HORIZONTAL, NULL, &narrow_nat);
 
     narrow_nat = MAX (narrow_nat, DIALOG_MIN_WIDTH);
 
-    if (max_size < wide_min)
+    if (max_size < wide_nat)
       max_size = MIN (max_size, narrow_nat);
   } else {
     max_size = priv->parent_height - DIALOG_MARGIN * 2;
@@ -826,10 +861,7 @@ adw_message_dialog_dispose (GObject *object)
 
   set_parent (self, NULL);
 
-  if (priv->child) {
-    gtk_box_remove (priv->message_area, priv->child);
-    priv->child = NULL;
-  }
+  priv->child = NULL;
 
   if (priv->responses) {
     g_list_free_full (priv->responses, (GDestroyNotify) response_info_free);
@@ -995,10 +1027,14 @@ adw_message_dialog_class_init (AdwMessageDialogClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                   G_STRUCT_OFFSET (AdwMessageDialogClass, response),
-                  NULL, NULL, NULL,
+                  NULL, NULL,
+                  adw_marshal_VOID__STRING,
                   G_TYPE_NONE,
                   1,
                   G_TYPE_STRING);
+  g_signal_set_va_marshaller (signals[SIGNAL_RESPONSE],
+                              G_TYPE_FROM_CLASS (klass),
+                              adw_marshal_VOID__STRINGv);
 
   gtk_widget_class_set_template_from_resource (widget_class,
                                                "/org/gnome/Adwaita/ui/adw-message-dialog.ui");
@@ -1376,6 +1412,11 @@ adw_message_dialog_set_heading (AdwMessageDialog *self,
   gtk_label_set_label (GTK_LABEL (priv->heading_label), heading);
   gtk_widget_set_visible (priv->heading_label, heading && *heading);
 
+  if (heading && *heading)
+    gtk_widget_add_css_class (GTK_WIDGET (priv->message_area), "has-heading");
+  else
+    gtk_widget_remove_css_class (GTK_WIDGET (priv->message_area), "has-heading");
+
   update_window_title (self);
 
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_HEADING]);
@@ -1582,6 +1623,11 @@ adw_message_dialog_set_body (AdwMessageDialog *self,
   gtk_label_set_label (GTK_LABEL (priv->body_label), body);
   gtk_widget_set_visible (priv->body_label, body && *body);
 
+  if (body && *body)
+    gtk_widget_add_css_class (GTK_WIDGET (priv->message_area), "has-body");
+  else
+    gtk_widget_remove_css_class (GTK_WIDGET (priv->message_area), "has-body");
+
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_BODY]);
 }
 
@@ -1772,6 +1818,10 @@ adw_message_dialog_set_extra_child (AdwMessageDialog *self,
   AdwMessageDialogPrivate *priv;
 
   g_return_if_fail (ADW_IS_MESSAGE_DIALOG (self));
+  g_return_if_fail (child == NULL || GTK_IS_WIDGET (child));
+
+  if (child)
+    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
 
   priv = adw_message_dialog_get_instance_private (self);
 
@@ -2308,6 +2358,8 @@ choose_response_cb (AdwMessageDialog *dialog,
   if (cancellable)
     g_signal_handlers_disconnect_by_func (cancellable, choose_cancelled_cb, task);
 
+  g_signal_handlers_disconnect_by_func (dialog, choose_response_cb, task);
+
   g_task_return_int (task, g_quark_from_string (response));
 
   g_object_unref (task);
