@@ -122,12 +122,23 @@ oscillate (AdwSpringAnimation *self,
    * for the differential equation m*ẍ+b*ẋ+kx = 0
    */
 
+  /* Critically damped */
+  /* DBL_EPSILON is too small for this specific comparison, so we use
+   * FLT_EPSILON even though it's doubles */
+  if (G_APPROX_VALUE (beta, omega0, FLT_EPSILON)) {
+    if (velocity)
+      *velocity = envelope * (-beta * t * v0 - beta * beta * t * x0 + v0);
+
+    return self->value_to + envelope * (x0 + (beta * x0 + v0) * t);
+  }
+
   /* Underdamped */
   if (beta < omega0) {
     double omega1 = sqrt ((omega0 * omega0) - (beta * beta));
 
     if (velocity)
       *velocity = envelope * (v0 * cos (omega1 * t) - (x0 * omega1 + (beta * beta * x0 + beta * v0) / (omega1)) * sin (omega1 * t));
+
     return self->value_to + envelope * (x0 * cos (omega1 * t) + ((beta * x0 + v0) / omega1) * sin (omega1 * t));
   }
 
@@ -137,13 +148,11 @@ oscillate (AdwSpringAnimation *self,
 
     if (velocity)
       *velocity = envelope * (v0 * coshl (omega2 * t) + (omega2 * x0 - (beta * beta * x0 + beta * v0) / omega2) * sinhl (omega2 * t));
+
     return self->value_to + envelope * (x0 * coshl (omega2 * t) + ((beta * x0 + v0) / omega2) * sinhl (omega2 * t));
   }
 
-  /* Critically damped */
-  if (velocity)
-    *velocity = envelope * (beta * x0 + v0) * (1 - beta);
-  return self->value_to + envelope * (x0 + (beta * x0 + v0) * t);
+  g_assert_not_reached ();
 }
 
 static guint
@@ -154,8 +163,8 @@ get_first_zero (AdwSpringAnimation *self)
   guint i = 1;
   double y = oscillate (self, i, NULL);
 
-  while ((self->value_to - self->value_from > FLT_EPSILON && self->value_to - y > self->epsilon) ||
-         (self->value_from - self->value_to > FLT_EPSILON && y - self->value_to > self->epsilon)) {
+  while ((self->value_to - self->value_from > DBL_EPSILON && self->value_to - y > self->epsilon) ||
+         (self->value_from - self->value_to > DBL_EPSILON && y - self->value_to > self->epsilon)) {
     if (i > MAX_ITERATIONS)
       return 0;
 
@@ -180,11 +189,11 @@ calculate_duration (AdwSpringAnimation *self)
 
   int i = 0;
 
-  if (beta <= 0)
+  if (G_APPROX_VALUE (beta, 0, DBL_EPSILON) || beta < 0)
     return ADW_DURATION_INFINITE;
 
   if (self->clamp) {
-    if (G_APPROX_VALUE (self->value_to, self->value_from, FLT_EPSILON))
+    if (G_APPROX_VALUE (self->value_to, self->value_from, DBL_EPSILON))
       return 0;
 
     return get_first_zero (self);
@@ -199,7 +208,9 @@ calculate_duration (AdwSpringAnimation *self)
    */
   x0 = -log (self->epsilon) / beta;
 
-  if (beta <= omega0)
+  /* DBL_EPSILON is too small for this specific comparison, so we use
+   * FLT_EPSILON even though it's doubles */
+  if (G_APPROX_VALUE (beta, omega0, FLT_EPSILON) || beta < omega0)
     return x0 * 1000;
 
   /*
@@ -208,22 +219,23 @@ calculate_duration (AdwSpringAnimation *self)
    * Newton's root finding method is a good candidate in this particular case:
    * https://en.wikipedia.org/wiki/Newton%27s_method
    */
-  y0 = oscillate (self, x0*1000, NULL);
+  y0 = oscillate (self, x0 * 1000, NULL);
   m = (oscillate (self, (x0 + DELTA) * 1000, NULL) - y0) / DELTA;
 
   x1 = (self->value_to - y0 + m * x0) / m;
-  y1 = oscillate (self, x1*1000, NULL);
+  y1 = oscillate (self, x1 * 1000, NULL);
 
   while (ABS (self->value_to - y1) > self->epsilon) {
     if (i>1000)
       return 0;
+
     x0 = x1;
     y0 = y1;
 
     m = (oscillate (self, (x0 + DELTA) * 1000, NULL) - y0) / DELTA;
 
     x1 = (self->value_to - y0 + m * x0) / m;
-    y1 = oscillate (self, x1*1000, NULL);
+    y1 = oscillate (self, x1 * 1000, NULL);
     i++;
   }
 
@@ -251,8 +263,8 @@ adw_spring_animation_estimate_duration (AdwAnimation *animation)
 }
 
 static double
-adw_spring_animation_calculate_value (AdwAnimation *animation,
-                                      guint         t)
+adw_spring_animation_real_calculate_value (AdwAnimation *animation,
+                                           guint         t)
 {
   AdwSpringAnimation *self = ADW_SPRING_ANIMATION (animation);
   double value;
@@ -386,7 +398,7 @@ adw_spring_animation_class_init (AdwSpringAnimationClass *klass)
   object_class->get_property = adw_spring_animation_get_property;
 
   animation_class->estimate_duration = adw_spring_animation_estimate_duration;
-  animation_class->calculate_value = adw_spring_animation_calculate_value;
+  animation_class->calculate_value = adw_spring_animation_real_calculate_value;
 
   /**
    * AdwSpringAnimation:value-from: (attributes org.gtk.Property.get=adw_spring_animation_get_value_from org.gtk.Property.set=adw_spring_animation_set_value_from)
@@ -590,7 +602,7 @@ adw_spring_animation_set_value_from (AdwSpringAnimation *self,
 {
   g_return_if_fail (ADW_IS_SPRING_ANIMATION (self));
 
-  if (G_APPROX_VALUE (self->value_from, value, FLT_EPSILON))
+  if (G_APPROX_VALUE (self->value_from, value, DBL_EPSILON))
     return;
 
   self->value_from = value;
@@ -632,7 +644,7 @@ adw_spring_animation_set_value_to (AdwSpringAnimation *self,
 {
   g_return_if_fail (ADW_IS_SPRING_ANIMATION (self));
 
-  if (G_APPROX_VALUE (self->value_to, value, FLT_EPSILON))
+  if (G_APPROX_VALUE (self->value_to, value, DBL_EPSILON))
     return;
 
   self->value_to = value;
@@ -714,7 +726,7 @@ adw_spring_animation_set_initial_velocity (AdwSpringAnimation *self,
 {
   g_return_if_fail (ADW_IS_SPRING_ANIMATION (self));
 
-  if (G_APPROX_VALUE (self->initial_velocity, velocity, FLT_EPSILON))
+  if (G_APPROX_VALUE (self->initial_velocity, velocity, DBL_EPSILON))
     return;
 
   self->initial_velocity = velocity;
@@ -765,7 +777,7 @@ adw_spring_animation_set_epsilon (AdwSpringAnimation *self,
   g_return_if_fail (ADW_IS_SPRING_ANIMATION (self));
   g_return_if_fail (epsilon> 0.0);
 
-  if (G_APPROX_VALUE (self->epsilon, epsilon, FLT_EPSILON))
+  if (G_APPROX_VALUE (self->epsilon, epsilon, DBL_EPSILON))
     return;
 
   self->epsilon = epsilon;
@@ -820,11 +832,65 @@ adw_spring_animation_set_clamp (AdwSpringAnimation *self,
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CLAMP]);
 }
 
+/**
+ * adw_spring_animation_calculate_value:
+ * @self: a spring animation
+ * @time: elapsed time, in milliseconds
+ *
+ * Calculates the value @self will have at @time.
+ *
+ * The time starts at 0 and ends at
+ * [property@SpringAnimation:estimated_duration].
+ *
+ * See also [method@SpringAnimation.calculate_velocity].
+ *
+ * Returns: the value at @time
+ *
+ * Since: 1.3
+ */
+double
+adw_spring_animation_calculate_value (AdwSpringAnimation *self,
+                                      guint               time)
+{
+  g_return_val_if_fail (ADW_IS_SPRING_ANIMATION (self), 0.0);
+
+  return oscillate (self, time, NULL);
+}
+
+/**
+ * adw_spring_animation_calculate_velocity:
+ * @self: a spring animation
+ * @time: elapsed time, in milliseconds
+ *
+ * Calculates the velocity @self will have at @time.
+ *
+ * The time starts at 0 and ends at
+ * [property@SpringAnimation:estimated_duration].
+ *
+ * See also [method@SpringAnimation.calculate_value].
+ *
+ * Returns: the velocity at @time
+ *
+ * Since: 1.3
+ */
+double
+adw_spring_animation_calculate_velocity (AdwSpringAnimation *self,
+                                         guint               time)
+{
+  double velocity;
+
+  g_return_val_if_fail (ADW_IS_SPRING_ANIMATION (self), 0.0);
+
+  oscillate (self, time, &velocity);
+
+  return velocity;
+}
+
 /**
  * adw_spring_animation_get_estimated_duration: (attributes org.gtk.Method.get_property=estimated-duration)
  * @self: a spring animation
  *
- * Gets the estimated duration of @self.
+ * Gets the estimated duration of @self, in milliseconds.
  *
  * Can be [const@DURATION_INFINITE] if the spring damping is set to 0.
  *
