@@ -3,7 +3,7 @@
  *
  * SPDX-License-Identifier: LGPL-2.1+
  *
- * Author: Alexander Mikhaylenko <alexander.mikhaylenko@puri.sm>
+ * Author: Alice Mikhaylenko <alice.mikhaylenko@puri.sm>
  */
 
 #include "config.h"
@@ -14,6 +14,7 @@
 #include "adw-animation-util.h"
 #include "adw-bin.h"
 #include "adw-header-bar.h"
+#include "adw-marshalers.h"
 #include "adw-style-manager.h"
 #include "adw-tab-grid-private.h"
 #include "adw-tab-thumbnail-private.h"
@@ -65,8 +66,12 @@
  * for the overview. Use it to add extra actions, e.g. to open a new window or
  * undo closed tab.
  *
- * `AdwTabOverview` is intended to be cover the whole window and shows window
- * buttons by default. They can be disabled by setting
+ * `AdwTabOverview` is intended to be used as the direct child of the window,
+ * with the rest of the window contents set as the [property@TabOverview:child].
+ * The child is expected to contain an [class@TabView].
+ *
+ * `AdwTabOverview` shows window buttons by default. They can be disabled by
+ * setting [property@TabOverview:show-start-title-buttons] and/or
  * [property@TabOverview:show-start-title-buttons] and/or
  * [property@TabOverview:show-end-title-buttons] to `FALSE`.
  *
@@ -120,6 +125,8 @@ struct _AdwTabOverview
   GtkWidget *transition_picture;
   gboolean transition_pinned;
 
+  GdkDragAction extra_drag_preferred_action;
+
   GtkWidget *last_focus;
 };
 
@@ -142,6 +149,7 @@ enum {
   PROP_SECONDARY_MENU,
   PROP_SHOW_START_TITLE_BUTTONS,
   PROP_SHOW_END_TITLE_BUTTONS,
+  PROP_EXTRA_DRAG_PREFERRED_ACTION,
   PROP_EXTRA_DRAG_PRELOAD,
   LAST_PROP
 };
@@ -523,11 +531,6 @@ adw_tab_overview_scrollable_size_allocate (GtkWidget *widget,
   /* The value may have changed during gtk_adjustment_configure() */
   value = floor (gtk_adjustment_get_value (self->vadjustment));
 
-  if (value <= 0)
-    gtk_widget_add_css_class (self->overview, "scrolled-to-top");
-  else
-    gtk_widget_remove_css_class (self->overview, "scrolled-to-top");
-
   adw_tab_grid_set_visible_range (ADW_TAB_GRID (self->pinned_grid),
                                   CLAMP (value - self->pinned_grid_pos, 0, pinned_height),
                                   CLAMP (value - self->pinned_grid_pos + height - new_button_height, 0, pinned_height),
@@ -720,14 +723,27 @@ adw_tab_overview_scrollable_init (AdwTabOverviewScrollable *self)
                             G_CALLBACK (scroll_animation_done_cb), self);
 }
 
+static void
+set_extra_drag_preferred_action (AdwTabOverview *self,
+                                 GdkDragAction   preferred_action)
+{
+  self->extra_drag_preferred_action = preferred_action;
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_EXTRA_DRAG_PREFERRED_ACTION]);
+}
+
+
 static gboolean
 extra_drag_drop_cb (AdwTabOverview *self,
                     AdwTabPage     *page,
-                    GValue         *value)
+                    GValue         *value,
+                    GdkDragAction   preferred_action)
 {
   gboolean ret = GDK_EVENT_PROPAGATE;
 
+  set_extra_drag_preferred_action (self, preferred_action);
   g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, page, value, &ret);
+  set_extra_drag_preferred_action (self, 0);
 
   return ret;
 }
@@ -744,6 +760,13 @@ extra_drag_value_cb (AdwTabOverview *self,
   return preferred_action;
 }
 
+static GdkDragAction
+extra_drag_value_notify (AdwTabOverview *self,
+                         GValue         *value)
+{
+  return GDK_ACTION_ALL;
+}
+
 static void
 empty_changed_cb (AdwTabOverview *self)
 {
@@ -891,6 +914,7 @@ notify_n_pages_cb (AdwTabOverview *self)
 
   n_pages = adw_tab_view_get_n_pages (self->view);
 
+  /* Translators: Tab overview title, %u is the number of open tabs */
   title_str = g_strdup_printf (ngettext ("%u Tab", "%u Tabs", n_pages), n_pages);
 
   adw_window_title_set_title (ADW_WINDOW_TITLE (self->title), title_str);
@@ -996,6 +1020,14 @@ open_animation_done_cb (AdwTabOverview *self)
   gtk_widget_queue_draw (GTK_WIDGET (self));
 }
 
+static inline double
+inverse_lerp (double a,
+              double b,
+              double t)
+{
+  return (t - a) / (b - a);
+}
+
 static void
 calculate_bounds (AdwTabOverview  *self,
                   graphene_rect_t *bounds,
@@ -1045,11 +1077,18 @@ calculate_bounds (AdwTabOverview  *self,
                       adw_lerp (1, thumbnail_bounds.size.width / view_bounds.size.width, self->progress),
                       adw_lerp (1, thumbnail_bounds.size.height / view_bounds.size.height, self->progress));
 
-  graphene_rect_init (transition_bounds,
-                      adw_lerp (0, thumbnail_bounds.origin.x, self->progress),
-                      adw_lerp (0, thumbnail_bounds.origin.y, self->progress),
+  graphene_size_init (&transition_bounds->size,
                       clip_bounds->size.width * clip_scale->width,
                       clip_bounds->size.height * clip_scale->height);
+  graphene_point_init (&transition_bounds->origin,
+                       adw_lerp (0, thumbnail_bounds.origin.x,
+                                 inverse_lerp (bounds->size.width,
+                                               thumbnail_bounds.size.width,
+                                               transition_bounds->size.width)),
+                       adw_lerp (0, thumbnail_bounds.origin.y,
+                                 inverse_lerp (bounds->size.height,
+                                               thumbnail_bounds.size.height,
+                                               transition_bounds->size.height)));
 }
 
 static void
@@ -1321,11 +1360,10 @@ adw_tab_overview_dispose (GObject *object)
 
   adw_tab_overview_set_view (self, NULL);
 
-  gtk_widget_unparent (GTK_WIDGET (self->overview));
-  gtk_widget_unparent (GTK_WIDGET (self->child_bin));
-
   g_clear_object (&self->open_animation);
 
+  gtk_widget_dispose_template (GTK_WIDGET (self), ADW_TYPE_TAB_OVERVIEW);
+
   G_OBJECT_CLASS (adw_tab_overview_parent_class)->dispose (object);
 }
 
@@ -1368,6 +1406,9 @@ adw_tab_overview_get_property (GObject    *object,
   case PROP_SHOW_END_TITLE_BUTTONS:
     g_value_set_boolean (value, adw_tab_overview_get_show_end_title_buttons (self));
     break;
+  case PROP_EXTRA_DRAG_PREFERRED_ACTION:
+    g_value_set_flags (value, adw_tab_overview_get_extra_drag_preferred_action (self));
+    break;
   case PROP_EXTRA_DRAG_PRELOAD:
     g_value_set_boolean (value, adw_tab_overview_get_extra_drag_preload (self));
     break;
@@ -1626,6 +1667,23 @@ adw_tab_overview_class_init (AdwTabOverviewClass *klass)
                           TRUE,
                           G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
+  /**
+   * AdwTabOverview:extra-drag-preferred-action: (attributes org.gtk.Property.get=adw_tab_overview_get_extra_drag_preferred_action)
+   *
+   * The unique action on the `current-drop` of the
+   * [signal@TabOverview::extra-drag-drop].
+   *
+   * This property should only be used during a
+   * [signal@TabOverview::extra-drag-drop] and is always a subset of what was
+   * originally passed to [method@TabOverview.setup_extra_drop_target].
+   *
+   * Since: 1.4
+   */
+  props[PROP_EXTRA_DRAG_PREFERRED_ACTION] =
+    g_param_spec_flags ("extra-drag-preferred-action", NULL, NULL,
+                       GDK_TYPE_DRAG_ACTION, 0,
+                       G_PARAM_READABLE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
   /**
    * AdwTabOverview:extra-drag-preload: (attributes org.gtk.Property.get=adw_tab_overview_get_extra_drag_preload org.gtk.Property.set=adw_tab_overview_set_extra_drag_preload)
    *
@@ -1664,9 +1722,13 @@ adw_tab_overview_class_init (AdwTabOverviewClass *klass)
                   G_SIGNAL_RUN_LAST,
                   0,
                   object_handled_accumulator,
-                  NULL, NULL,
+                  NULL,
+                  adw_marshal_OBJECT__VOID,
                   ADW_TYPE_TAB_PAGE,
                   0);
+  g_signal_set_va_marshaller (signals[SIGNAL_CREATE_TAB],
+                              G_TYPE_FROM_CLASS (klass),
+                              adw_marshal_OBJECT__VOIDv);
 
   /**
    * AdwTabOverview::extra-drag-drop:
@@ -1690,7 +1752,8 @@ adw_tab_overview_class_init (AdwTabOverviewClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  g_signal_accumulator_first_wins, NULL, NULL,
+                  g_signal_accumulator_first_wins,
+                  NULL, NULL,
                   G_TYPE_BOOLEAN,
                   2,
                   ADW_TYPE_TAB_PAGE,
@@ -1721,7 +1784,8 @@ adw_tab_overview_class_init (AdwTabOverviewClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  g_signal_accumulator_first_wins, NULL, NULL,
+                  g_signal_accumulator_first_wins,
+                  NULL, NULL,
                   GDK_TYPE_DRAG_ACTION,
                   2,
                   ADW_TYPE_TAB_PAGE,
@@ -1761,6 +1825,9 @@ adw_tab_overview_class_init (AdwTabOverviewClass *klass)
   gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
   gtk_widget_class_set_css_name (widget_class, "taboverview");
 
+  g_signal_override_class_handler ("extra-drag-value", G_TYPE_FROM_CLASS (klass),
+                                   G_CALLBACK (extra_drag_value_notify));
+
   g_type_ensure (ADW_TYPE_TAB_GRID);
   g_type_ensure (ADW_TYPE_TAB_OVERVIEW_SCROLLABLE);
 }
@@ -1961,6 +2028,9 @@ adw_tab_overview_set_child (AdwTabOverview *self,
   g_return_if_fail (ADW_IS_TAB_OVERVIEW (self));
   g_return_if_fail (child == NULL || GTK_IS_WIDGET (child));
 
+  if (child)
+    g_return_if_fail (gtk_widget_get_parent (child) == NULL);
+
   if (child == adw_tab_overview_get_child (self))
     return;
 
@@ -2454,6 +2524,24 @@ adw_tab_overview_get_pinned_tab_grid (AdwTabOverview *self)
   return self->pinned_grid;
 }
 
+/**
+ * adw_tab_overview_get_extra_drag_preferred_action:
+ * @self: a tab overview
+ *
+ * Gets the current action during a drop on the extra_drop_target.
+ *
+ * Returns: the drag action of the current drop.
+ *
+ * Since: 1.4
+ */
+GdkDragAction
+adw_tab_overview_get_extra_drag_preferred_action (AdwTabOverview *self)
+{
+  g_return_val_if_fail (ADW_IS_TAB_OVERVIEW (self), 0);
+
+  return self->extra_drag_preferred_action;
+}
+
 /**
  * adw_tab_overview_get_extra_drag_preload: (attributes org.gtk.Method.get_property=extra-drag-preload)
  * @self: a tab overview
