@@ -3,7 +3,7 @@
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  *
- * Author: Alexander Mikhaylenko <alexander.mikhaylenko@puri.sm>
+ * Author: Alice Mikhaylenko <alice.mikhaylenko@puri.sm>
  */
 
 #include "config.h"
@@ -193,17 +193,45 @@ indicator_clicked_cb (AdwTabThumbnail *self)
   g_signal_emit_by_name (self->view, "indicator-activated", self->page);
 }
 
+static GdkDragAction
+make_action_unique (GdkDragAction actions)
+{
+  if (actions & GDK_ACTION_COPY)
+    return GDK_ACTION_COPY;
+
+  if (actions & GDK_ACTION_MOVE)
+    return GDK_ACTION_MOVE;
+
+  if (actions & GDK_ACTION_LINK)
+    return GDK_ACTION_LINK;
+
+  return 0;
+}
+
 static gboolean
 drop_cb (AdwTabThumbnail *self,
          GValue          *value)
 {
   gboolean ret = GDK_EVENT_PROPAGATE;
+  GdkDrop *drop = gtk_drop_target_get_current_drop (self->drop_target);
+  GdkDragAction preferred_action = gdk_drop_get_actions (drop);
 
-  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, &ret);
+  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, preferred_action, &ret);
 
   return ret;
 }
 
+static GdkDragAction
+extra_drag_enter_cb (AdwTabThumbnail *self)
+{
+  const GValue *value = gtk_drop_target_get_value (self->drop_target);
+
+  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_VALUE], 0, value, &self->preferred_action);
+  self->preferred_action = make_action_unique (self->preferred_action);
+
+  return self->preferred_action;
+}
+
 static GdkDragAction
 extra_drag_motion_cb (AdwTabThumbnail *self)
 {
@@ -216,21 +244,7 @@ extra_drag_notify_value_cb (AdwTabThumbnail *self)
   const GValue *value = gtk_drop_target_get_value (self->drop_target);
 
   g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_VALUE], 0, value, &self->preferred_action);
-}
-
-static GdkDragAction
-make_action_unique (GdkDragAction actions)
-{
-  if (actions & GDK_ACTION_COPY)
-    return GDK_ACTION_COPY;
-
-  if (actions & GDK_ACTION_MOVE)
-    return GDK_ACTION_MOVE;
-
-  if (actions & GDK_ACTION_LINK)
-    return GDK_ACTION_LINK;
-
-  return 0;
+  self->preferred_action = make_action_unique (self->preferred_action);
 }
 
 static void
@@ -402,8 +416,7 @@ adw_tab_thumbnail_dispose (GObject *object)
 
   g_clear_object (&self->fade_animation);
 
-  if (self->contents)
-    gtk_widget_unparent (self->contents);
+  gtk_widget_dispose_template (GTK_WIDGET (self), ADW_TYPE_TAB_THUMBNAIL);
 
   G_OBJECT_CLASS (adw_tab_thumbnail_parent_class)->dispose (object);
 }
@@ -509,17 +522,20 @@ adw_tab_thumbnail_class_init (AdwTabThumbnailClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  g_signal_accumulator_first_wins, NULL, NULL,
+                  g_signal_accumulator_first_wins,
+                  NULL, NULL,
                   G_TYPE_BOOLEAN,
-                  1,
-                  G_TYPE_VALUE);
+                  2,
+                  G_TYPE_VALUE,
+                  GDK_TYPE_DRAG_ACTION);
 
   signals[SIGNAL_EXTRA_DRAG_VALUE] =
     g_signal_new ("extra-drag-value",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  g_signal_accumulator_first_wins, NULL, NULL,
+                  g_signal_accumulator_first_wins,
+                  NULL, NULL,
                   GDK_TYPE_DRAG_ACTION,
                   1,
                   G_TYPE_VALUE);
@@ -543,6 +559,7 @@ adw_tab_thumbnail_class_init (AdwTabThumbnailClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, unpin_clicked_cb);
   gtk_widget_class_bind_template_callback (widget_class, indicator_clicked_cb);
   gtk_widget_class_bind_template_callback (widget_class, drop_cb);
+  gtk_widget_class_bind_template_callback (widget_class, extra_drag_enter_cb);
   gtk_widget_class_bind_template_callback (widget_class, extra_drag_motion_cb);
   gtk_widget_class_bind_template_callback (widget_class, extra_drag_notify_value_cb);
 
@@ -718,6 +735,8 @@ adw_tab_thumbnail_fade_out (AdwTabThumbnail *self)
   g_return_if_fail (ADW_IS_TAB_THUMBNAIL (self));
 
   adw_animation_reset (self->fade_animation);
+
+  gtk_widget_set_opacity (self->overlay, 0);
 }
 
 void
@@ -725,5 +744,7 @@ adw_tab_thumbnail_fade_in (AdwTabThumbnail *self)
 {
   g_return_if_fail (ADW_IS_TAB_THUMBNAIL (self));
 
+  gtk_widget_set_opacity (self->overlay, 1);
+
   adw_animation_play (self->fade_animation);
 }
