@@ -3,7 +3,7 @@
  *
  * SPDX-License-Identifier: LGPL-2.1-or-later
  *
- * Author: Alexander Mikhaylenko <alexander.mikhaylenko@puri.sm>
+ * Author: Alice Mikhaylenko <alice.mikhaylenko@puri.sm>
  */
 
 #include "config.h"
@@ -56,9 +56,6 @@ struct _AdwTab
 
   AdwAnimation *close_btn_animation;
   AdwAnimation *needs_attention_animation;
-
-  GskGLShader *shader;
-  gboolean shader_compiled;
 };
 
 G_DEFINE_FINAL_TYPE (AdwTab, adw_tab, GTK_TYPE_WIDGET)
@@ -275,6 +272,21 @@ indicator_clicked_cb (AdwTab *self)
   g_signal_emit_by_name (self->view, "indicator-activated", self->page);
 }
 
+static GdkDragAction
+make_action_unique (GdkDragAction actions)
+{
+  if (actions & GDK_ACTION_COPY)
+    return GDK_ACTION_COPY;
+
+  if (actions & GDK_ACTION_MOVE)
+    return GDK_ACTION_MOVE;
+
+  if (actions & GDK_ACTION_LINK)
+    return GDK_ACTION_LINK;
+
+  return 0;
+}
+
 static void
 enter_cb (AdwTab             *self,
           double              x,
@@ -320,12 +332,25 @@ drop_cb (AdwTab *self,
          GValue *value)
 {
   gboolean ret = GDK_EVENT_PROPAGATE;
+  GdkDrop *drop = gtk_drop_target_get_current_drop (self->drop_target);
+  GdkDragAction preferred_action = gdk_drop_get_actions (drop);
 
-  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, &ret);
+  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_DROP], 0, value, preferred_action, &ret);
 
   return ret;
 }
 
+static GdkDragAction
+extra_drag_enter_cb (AdwTab *self)
+{
+  const GValue *value = gtk_drop_target_get_value (self->drop_target);
+
+  g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_VALUE], 0, value, &self->preferred_action);
+  self->preferred_action = make_action_unique (self->preferred_action);
+
+  return self->preferred_action;
+}
+
 static GdkDragAction
 extra_drag_motion_cb (AdwTab *self)
 {
@@ -338,48 +363,7 @@ extra_drag_notify_value_cb (AdwTab *self)
   const GValue *value = gtk_drop_target_get_value (self->drop_target);
 
   g_signal_emit (self, signals[SIGNAL_EXTRA_DRAG_VALUE], 0, value, &self->preferred_action);
-}
-
-static GdkDragAction
-make_action_unique (GdkDragAction actions)
-{
-  if (actions & GDK_ACTION_COPY)
-    return GDK_ACTION_COPY;
-
-  if (actions & GDK_ACTION_MOVE)
-    return GDK_ACTION_MOVE;
-
-  if (actions & GDK_ACTION_LINK)
-    return GDK_ACTION_LINK;
-
-  return 0;
-}
-
-static void
-ensure_shader (AdwTab *self)
-{
-  GtkNative *native;
-  GskRenderer *renderer;
-  GError *error = NULL;
-
-  if (self->shader)
-    return;
-
-  self->shader = gsk_gl_shader_new_from_resource ("/org/gnome/Adwaita/glsl/fade.glsl");
-
-  native = gtk_widget_get_native (GTK_WIDGET (self));
-  renderer = gtk_native_get_renderer (native);
-
-  self->shader_compiled = gsk_gl_shader_compile (self->shader, renderer, &error);
-
-  if (error) {
-    /* If shaders aren't supported, the error doesn't matter and we just
-     * silently fall back */
-    if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED))
-      g_warning ("Couldn't compile shader: %s\n", error->message);
-  }
-
-  g_clear_error (&error);
+  self->preferred_action = make_action_unique (self->preferred_action);
 }
 
 static gboolean
@@ -621,43 +605,32 @@ adw_tab_snapshot (GtkWidget   *widget,
     gboolean is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
     int width = gtk_widget_get_width (widget);
     int height = gtk_widget_get_height (widget);
-    float offset =
-      gtk_widget_get_allocated_width (self->close_btn) +
-      gtk_widget_get_margin_end (self->title);
-    graphene_rect_t bounds;
+    float offset = gtk_widget_get_width (self->close_btn) +
+                   gtk_widget_get_margin_end (self->title);
 
-    ensure_shader (self);
+    gtk_snapshot_push_mask (snapshot, GSK_MASK_MODE_INVERTED_ALPHA);
 
-    graphene_rect_init (&bounds, 0, 0, width, height);
-
-    if (self->shader_compiled) {
-      gtk_snapshot_push_gl_shader (snapshot, self->shader, &bounds,
-                                   gsk_gl_shader_format_args (self->shader,
-                                                              "offsetLeft", is_rtl ? offset : 0.0f,
-                                                              "offsetRight", is_rtl ? 0.0f : offset,
-                                                              "strengthLeft", is_rtl ? opacity : 0.0f,
-                                                              "strengthRight", is_rtl ? 0.0f : opacity,
-                                                              "widthLeft", FADE_WIDTH,
-                                                              "widthRight", FADE_WIDTH,
-                                                              NULL));
-    } else {
-      bounds.size.width -= offset;
-
-      if (is_rtl)
-        bounds.origin.x += offset;
-
-      gtk_snapshot_push_clip (snapshot, &bounds);
+    if (!is_rtl) {
+      gtk_snapshot_translate (snapshot, &GRAPHENE_POINT_INIT (width, 0));
+      gtk_snapshot_scale (snapshot, -1, 1);
     }
+
+    gtk_snapshot_append_linear_gradient (snapshot,
+                                         &GRAPHENE_RECT_INIT (0, 0, FADE_WIDTH + offset, height),
+                                         &GRAPHENE_POINT_INIT (offset, 0),
+                                         &GRAPHENE_POINT_INIT (FADE_WIDTH + offset, 0),
+                                         (GskColorStop[2]) {
+                                             { 0, { 0, 0, 0, opacity } },
+                                             { 1, { 0, 0, 0, 0 } },
+                                         },
+                                         2);
+    gtk_snapshot_pop (snapshot);
   }
 
   gtk_widget_snapshot_child (widget, self->title, snapshot);
 
-  if (draw_fade) {
-    if (self->shader_compiled)
-      gtk_snapshot_gl_shader_pop_texture (snapshot);
-
+  if (draw_fade)
     gtk_snapshot_pop (snapshot);
-  }
 
   gtk_widget_snapshot_child (widget, self->close_btn, snapshot);
 }
@@ -674,16 +647,6 @@ adw_tab_direction_changed (GtkWidget        *widget,
                                                               previous_direction);
 }
 
-static void
-adw_tab_unrealize (GtkWidget *widget)
-{
-  AdwTab *self = ADW_TAB (widget);
-
-  GTK_WIDGET_CLASS (adw_tab_parent_class)->unrealize (widget);
-
-  g_clear_object (&self->shader);
-}
-
 static void
 adw_tab_constructed (GObject *object)
 {
@@ -779,14 +742,10 @@ adw_tab_dispose (GObject *object)
 
   adw_tab_set_page (self, NULL);
 
-  g_clear_object (&self->shader);
   g_clear_object (&self->close_btn_animation);
   g_clear_object (&self->needs_attention_animation);
-  gtk_widget_unparent (self->indicator_btn);
-  gtk_widget_unparent (self->icon_stack);
-  gtk_widget_unparent (self->title);
-  gtk_widget_unparent (self->close_btn);
-  gtk_widget_unparent (self->needs_attention_indicator);
+
+  gtk_widget_dispose_template (GTK_WIDGET (self), ADW_TYPE_TAB);
 
   G_OBJECT_CLASS (adw_tab_parent_class)->dispose (object);
 }
@@ -808,7 +767,6 @@ adw_tab_class_init (AdwTabClass *klass)
   widget_class->unmap = adw_tab_unmap;
   widget_class->snapshot = adw_tab_snapshot;
   widget_class->direction_changed = adw_tab_direction_changed;
-  widget_class->unrealize = adw_tab_unrealize;
 
   props[PROP_VIEW] =
     g_param_spec_object ("view", NULL, NULL,
@@ -840,17 +798,20 @@ adw_tab_class_init (AdwTabClass *klass)
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  g_signal_accumulator_first_wins, NULL, NULL,
+                  g_signal_accumulator_first_wins,
+                  NULL, NULL,
                   G_TYPE_BOOLEAN,
-                  1,
-                  G_TYPE_VALUE);
+                  2,
+                  G_TYPE_VALUE,
+                  GDK_TYPE_DRAG_ACTION);
 
   signals[SIGNAL_EXTRA_DRAG_VALUE] =
     g_signal_new ("extra-drag-value",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_LAST,
                   0,
-                  g_signal_accumulator_first_wins, NULL, NULL,
+                  g_signal_accumulator_first_wins,
+                  NULL, NULL,
                   GDK_TYPE_DRAG_ACTION,
                   1,
                   G_TYPE_VALUE);
@@ -874,6 +835,7 @@ adw_tab_class_init (AdwTabClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, motion_cb);
   gtk_widget_class_bind_template_callback (widget_class, leave_cb);
   gtk_widget_class_bind_template_callback (widget_class, drop_cb);
+  gtk_widget_class_bind_template_callback (widget_class, extra_drag_enter_cb);
   gtk_widget_class_bind_template_callback (widget_class, extra_drag_motion_cb);
   gtk_widget_class_bind_template_callback (widget_class, extra_drag_notify_value_cb);
 
@@ -884,6 +846,7 @@ adw_tab_class_init (AdwTabClass *klass)
   gtk_widget_class_add_binding (widget_class, GDK_KEY_KP_Enter,  0, (GtkShortcutFunc) activate_cb, NULL);
 
   gtk_widget_class_set_css_name (widget_class, "tab");
+  gtk_widget_class_set_accessible_role (widget_class, GTK_ACCESSIBLE_ROLE_TAB);
 
   g_type_ensure (ADW_TYPE_FADING_LABEL);
   g_type_ensure (ADW_TYPE_GIZMO);
