@@ -8,8 +8,7 @@
 
 #include "adw-toast-private.h"
 
-#include "adw-macros-private.h"
-
+#include "adw-marshalers.h"
 #include <stdarg.h>
 
 /**
@@ -20,8 +19,6 @@
  *   the previous toast into the queue instead.
  *
  * [class@Toast] behavior when another toast is already displayed.
- *
- * Since: 1.0
  */
 
 /**
@@ -48,6 +45,9 @@
  * [property@Toast:priority] determines how it behaves if another toast is
  * already being displayed.
  *
+ * Toast titles use Pango markup by default, set [property@Toast:use-markup] to
+ * `FALSE` if this is unwanted.
+ *
  * [property@Toast:custom-title] can be used to replace the title label with a
  * custom widget.
  *
@@ -145,8 +145,6 @@
  *   <source srcset="toast-undo-dark.png" media="(prefers-color-scheme: dark)">
  *   <img src="toast-undo.png" alt="toast-undo">
  * </picture>
- *
- * Since: 1.0
  */
 
 struct _AdwToast {
@@ -159,6 +157,7 @@ struct _AdwToast {
   AdwToastPriority priority;
   guint timeout;
   GtkWidget *custom_title;
+  gboolean use_markup;
 
   AdwToastOverlay *overlay;
 };
@@ -172,13 +171,15 @@ enum {
   PROP_PRIORITY,
   PROP_TIMEOUT,
   PROP_CUSTOM_TITLE,
-  LAST_PROP,
+  PROP_USE_MARKUP,
+  LAST_PROP
 };
 
 static GParamSpec *props[LAST_PROP];
 
 enum {
   SIGNAL_DISMISSED,
+  SIGNAL_BUTTON_CLICKED,
   SIGNAL_LAST_SIGNAL,
 };
 
@@ -245,6 +246,9 @@ adw_toast_get_property (GObject    *object,
   case PROP_CUSTOM_TITLE:
     g_value_set_object (value, adw_toast_get_custom_title (self));
     break;
+  case PROP_USE_MARKUP:
+    g_value_set_boolean (value, adw_toast_get_use_markup (self));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -280,6 +284,9 @@ adw_toast_set_property (GObject      *object,
   case PROP_CUSTOM_TITLE:
     adw_toast_set_custom_title (self, g_value_get_object (value));
     break;
+  case PROP_USE_MARKUP:
+    adw_toast_set_use_markup (self, g_value_get_boolean (value));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
   }
@@ -305,8 +312,6 @@ adw_toast_class_init (AdwToastClass *klass)
    * Setting a title will unset [property@Toast:custom-title].
    *
    * If [property@Toast:custom-title] is set, it will be used instead.
-   *
-   * Since: 1.0
    */
   props[PROP_TITLE] =
     g_param_spec_string ("title", NULL, NULL,
@@ -323,8 +328,6 @@ adw_toast_class_init (AdwToastClass *klass)
    * If set to `NULL`, the button won't be shown.
    *
    * See [property@Toast:action-name].
-   *
-   * Since: 1.0
    */
   props[PROP_BUTTON_LABEL] =
     g_param_spec_string ("button-label", NULL, NULL,
@@ -339,8 +342,6 @@ adw_toast_class_init (AdwToastClass *klass)
    * It will be activated when clicking the button.
    *
    * See [property@Toast:action-target].
-   *
-   * Since: 1.0
    */
   props[PROP_ACTION_NAME] =
     g_param_spec_string ("action-name", NULL, NULL,
@@ -351,10 +352,6 @@ adw_toast_class_init (AdwToastClass *klass)
    * AdwToast:action-target: (attributes org.gtk.Property.get=adw_toast_get_action_target_value org.gtk.Property.set=adw_toast_set_action_target_value)
    *
    * The parameter for action invocations.
-   *
-   * See [property@Toast:action-name].
-   *
-   * Since: 1.0
    */
   props[PROP_ACTION_TARGET] =
     g_param_spec_variant ("action-target", NULL, NULL,
@@ -374,8 +371,6 @@ adw_toast_class_init (AdwToastClass *klass)
    *
    * If the priority is `ADW_TOAST_PRIORITY_HIGH`, the toast will be displayed
    * immediately, pushing the previous toast into the queue instead.
-   *
-   * Since: 1.0
    */
   props[PROP_PRIORITY] =
     g_param_spec_enum ("priority", NULL, NULL,
@@ -393,8 +388,6 @@ adw_toast_class_init (AdwToastClass *klass)
    *
    * Toasts cannot disappear while being hovered, pressed (on touchscreen), or
    * have keyboard focus inside them.
-   *
-   * Since: 1.0
    */
   props[PROP_TIMEOUT] =
     g_param_spec_uint ("timeout", NULL, NULL,
@@ -418,22 +411,59 @@ adw_toast_class_init (AdwToastClass *klass)
                          GTK_TYPE_WIDGET,
                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
+  /**
+   * AdwToast:use-markup: (attributes org.gtk.Property.get=adw_toast_get_use_markup org.gtk.Property.set=adw_toast_set_use_markup)
+   *
+   * Whether to use Pango markup for the toast title.
+   *
+   * See also [func@Pango.parse_markup].
+   *
+   * Since: 1.4
+   */
+  props[PROP_USE_MARKUP] =
+    g_param_spec_boolean ("use-markup", NULL, NULL,
+                          TRUE,
+                          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
   /**
    * AdwToast::dismissed:
    *
    * Emitted when the toast has been dismissed.
-   *
-   * Since: 1.0
    */
   signals[SIGNAL_DISMISSED] =
     g_signal_new ("dismissed",
                   G_TYPE_FROM_CLASS (klass),
                   G_SIGNAL_RUN_FIRST,
                   0,
-                  NULL, NULL, NULL,
+                  NULL, NULL,
+                  adw_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
+  g_signal_set_va_marshaller (signals[SIGNAL_DISMISSED],
+                              G_TYPE_FROM_CLASS (klass),
+                              adw_marshal_VOID__VOIDv);
+
+  /**
+   * AdwToast::button-clicked:
+   *
+   * Emitted after the button has been clicked.
+   *
+   * It can be used as an alternative to setting an action.
+   *
+   * Since: 1.2
+   */
+  signals[SIGNAL_BUTTON_CLICKED] =
+    g_signal_new ("button-clicked",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL,
+                  adw_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
+  g_signal_set_va_marshaller (signals[SIGNAL_BUTTON_CLICKED],
+                              G_TYPE_FROM_CLASS (klass),
+                              adw_marshal_VOID__VOIDv);
 }
 
 static void
@@ -443,6 +473,7 @@ adw_toast_init (AdwToast *self)
   self->priority = ADW_TOAST_PRIORITY_NORMAL;
   self->timeout = 5;
   self->custom_title = NULL;
+  self->use_markup = TRUE;
 
   g_signal_connect (self, "dismissed", G_CALLBACK (dismissed_cb), self);
 }
@@ -458,8 +489,6 @@ adw_toast_init (AdwToast *self)
  * @title can be marked up with the Pango text markup language.
  *
  * Returns: the new created `AdwToast`
- *
- * Since: 1.0
  */
 AdwToast *
 adw_toast_new (const char *title)
@@ -517,8 +546,6 @@ adw_toast_new_format (const char *format,
  * the return value will be %NULL.
  *
  * Returns: (nullable): the title
- *
- * Since: 1.0
  */
 const char *
 adw_toast_get_title (AdwToast *self)
@@ -538,7 +565,11 @@ adw_toast_get_title (AdwToast *self)
  *
  * Sets the title that will be displayed on the toast.
  *
- * Since: 1.0
+ * The title can be marked up with the Pango text markup language.
+ *
+ * Setting a title will unset [property@Toast:custom-title].
+ *
+ * If [property@Toast:custom-title] is set, it will be used instead.
  */
 void
 adw_toast_set_title (AdwToast   *self,
@@ -569,8 +600,6 @@ adw_toast_set_title (AdwToast   *self,
  * Gets the label to show on the button.
  *
  * Returns: (nullable): the button label
- *
- * Since: 1.0
  */
 const char *
 adw_toast_get_button_label (AdwToast *self)
@@ -587,9 +616,11 @@ adw_toast_get_button_label (AdwToast *self)
  *
  * Sets the label to show on the button.
  *
- * It set to `NULL`, the button won't be shown.
+ * Underlines in the button text can be used to indicate a mnemonic.
  *
- * Since: 1.0
+ * If set to `NULL`, the button won't be shown.
+ *
+ * See [property@Toast:action-name].
  */
 void
 adw_toast_set_button_label (AdwToast   *self,
@@ -613,8 +644,6 @@ adw_toast_set_button_label (AdwToast   *self,
  * Gets the name of the associated action.
  *
  * Returns: (nullable): the action name
- *
- * Since: 1.0
  */
 const char *
 adw_toast_get_action_name (AdwToast *self)
@@ -631,7 +660,9 @@ adw_toast_get_action_name (AdwToast *self)
  *
  * Sets the name of the associated action.
  *
- * Since: 1.0
+ * It will be activated when clicking the button.
+ *
+ * See [property@Toast:action-target].
  */
 void
 adw_toast_set_action_name (AdwToast   *self,
@@ -655,8 +686,6 @@ adw_toast_set_action_name (AdwToast   *self,
  * Gets the parameter for action invocations.
  *
  * Returns: (transfer none) (nullable): the action target
- *
- * Since: 1.0
  */
 GVariant *
 adw_toast_get_action_target_value (AdwToast *self)
@@ -675,8 +704,6 @@ adw_toast_get_action_target_value (AdwToast *self)
  *
  * If the @action_target variant has a floating reference this function
  * will sink it.
- *
- * Since: 1.0
  */
 void
 adw_toast_set_action_target_value (AdwToast *self,
@@ -713,8 +740,6 @@ adw_toast_set_action_target_value (AdwToast *self,
  * If you are setting a string-valued target and want to set
  * the action name at the same time, you can use
  * [method@Toast.set_detailed_action_name].
-
- * Since: 1.0
  */
 void
 adw_toast_set_action_target (AdwToast   *self,
@@ -739,8 +764,6 @@ adw_toast_set_action_target (AdwToast   *self,
  *
  * @detailed_action_name is a string in the format accepted by
  * [func@Gio.Action.parse_detailed_name].
- *
- * Since: 1.0
  */
 void
 adw_toast_set_detailed_action_name (AdwToast   *self,
@@ -778,8 +801,6 @@ adw_toast_set_detailed_action_name (AdwToast   *self,
  * Gets priority for @self.
  *
  * Returns: the priority
- *
- * Since: 1.0
  */
 AdwToastPriority
 adw_toast_get_priority (AdwToast *self)
@@ -801,10 +822,8 @@ adw_toast_get_priority (AdwToast *self)
  *
  * If @priority is `ADW_TOAST_PRIORITY_NORMAL`, the toast will be queued.
  *
- * If @priority is `ADW_TOAST_PRIORITY_HIGH`, the toast will be displayed immediately,
- * pushing the previous toast into the queue instead.
- *
- * Since: 1.0
+ * If @priority is `ADW_TOAST_PRIORITY_HIGH`, the toast will be displayed
+ * immediately, pushing the previous toast into the queue instead.
  */
 void
 adw_toast_set_priority (AdwToast         *self,
@@ -829,8 +848,6 @@ adw_toast_set_priority (AdwToast         *self,
  * Gets timeout for @self.
  *
  * Returns: the timeout
- *
- * Since: 1.0
  */
 guint
 adw_toast_get_timeout (AdwToast *self)
@@ -852,8 +869,6 @@ adw_toast_get_timeout (AdwToast *self)
  *
  * Toasts cannot disappear while being hovered, pressed (on touchscreen), or
  * have keyboard focus inside them.
- *
- * Since: 1.0
  */
 void
 adw_toast_set_timeout (AdwToast *self,
@@ -869,58 +884,6 @@ adw_toast_set_timeout (AdwToast *self,
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_TIMEOUT]);
 }
 
-/**
- * adw_toast_dismiss:
- * @self: a toast
- *
- * Dismisses @self.
- *
- * Does nothing if @self has already been dismissed, or hasn't been added to an
- * [class@ToastOverlay].
- *
- * Since: 1.0
- */
-void
-adw_toast_dismiss (AdwToast *self)
-{
-  g_return_if_fail (ADW_IS_TOAST (self));
-
-  if (!self->overlay)
-    return;
-
-  g_signal_emit (self, signals[SIGNAL_DISMISSED], 0, NULL);
-}
-
-/**
- * adw_toast_set_custom_title: (attributes org.gtk.Method.set_property=custom-title)
- * @self: a toast
- * @widget: (nullable): the custom title widget
- *
- * Sets the custom title widget of @self.
- *
- * Since: 1.2
- */
-void
-adw_toast_set_custom_title (AdwToast  *self,
-                            GtkWidget *widget)
-{
-  g_return_if_fail (ADW_IS_TOAST (self));
-  g_return_if_fail (widget == NULL || GTK_IS_WIDGET (widget));
-
-  if (self->custom_title == widget)
-    return;
-
-  g_object_freeze_notify (G_OBJECT (self));
-
-  adw_toast_set_title (self, "");
-
-  g_set_object (&self->custom_title, widget);
-
-  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CUSTOM_TITLE]);
-
-  g_object_thaw_notify (G_OBJECT (self));
-}
-
 /**
  * adw_toast_get_custom_title: (attributes org.gtk.Method.get_property=custom-title)
  * @self: a toast
@@ -939,6 +902,64 @@ adw_toast_get_custom_title (AdwToast *self)
   return self->custom_title;
 }
 
+/**
+ * adw_toast_set_custom_title: (attributes org.gtk.Method.set_property=custom-title)
+ * @self: a toast
+ * @widget: (nullable): the custom title widget
+ *
+ * Sets the custom title widget of @self.
+ *
+ * It will be displayed instead of the title if set. In this case,
+ * [property@Toast:title] is ignored.
+ *
+ * Setting a custom title will unset [property@Toast:title].
+ *
+ * Since: 1.2
+ */
+void
+adw_toast_set_custom_title (AdwToast  *self,
+                            GtkWidget *widget)
+{
+  g_return_if_fail (ADW_IS_TOAST (self));
+  g_return_if_fail (widget == NULL || GTK_IS_WIDGET (widget));
+
+  if (widget)
+    g_return_if_fail (gtk_widget_get_parent (widget) == NULL);
+
+  if (self->custom_title == widget)
+    return;
+
+  g_object_freeze_notify (G_OBJECT (self));
+
+  adw_toast_set_title (self, "");
+
+  g_set_object (&self->custom_title, widget);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_CUSTOM_TITLE]);
+
+  g_object_thaw_notify (G_OBJECT (self));
+}
+
+/**
+ * adw_toast_dismiss:
+ * @self: a toast
+ *
+ * Dismisses @self.
+ *
+ * Does nothing if @self has already been dismissed, or hasn't been added to an
+ * [class@ToastOverlay].
+ */
+void
+adw_toast_dismiss (AdwToast *self)
+{
+  g_return_if_fail (ADW_IS_TOAST (self));
+
+  if (!self->overlay)
+    return;
+
+  g_signal_emit (self, signals[SIGNAL_DISMISSED], 0, NULL);
+}
+
 AdwToastOverlay *
 adw_toast_get_overlay (AdwToast *self)
 {
@@ -956,3 +977,48 @@ adw_toast_set_overlay (AdwToast        *self,
 
   self->overlay = overlay;
 }
+
+/**
+ * adw_toast_get_use_markup: (attributes org.gtk.Method.get_property=use-markup)
+ * @self: a toast
+ *
+ * Gets whether to use Pango markup for the toast title.
+ *
+ * Returns: whether the toast uses markup
+ *
+ * Since: 1.4
+ */
+gboolean
+adw_toast_get_use_markup (AdwToast *self)
+{
+  g_return_val_if_fail (ADW_IS_TOAST (self), FALSE);
+
+  return self->use_markup;
+}
+
+/**
+ * adw_toast_set_use_markup: (attributes org.gtk.Method.set_property=use-markup)
+ * @self: a toast
+ * @use_markup: whether to use markup
+ *
+ * Whether to use Pango markup for the toast title.
+ *
+ * See also [func@Pango.parse_markup].
+ *
+ * Since: 1.4
+ */
+void
+adw_toast_set_use_markup (AdwToast *self,
+                          gboolean  use_markup)
+{
+  g_return_if_fail (ADW_IS_TOAST (self));
+
+  use_markup = !!use_markup;
+
+  if (adw_toast_get_use_markup (self) == use_markup)
+    return;
+
+  self->use_markup = use_markup;
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_USE_MARKUP]);
+}
